/*
 * Copyright (C) 2012, Thibaud Briand, ENS Cachan <thibaud.briand@ens-cachan.fr>
 * Copyright (C) 2012, Jonathan Vacher, ENS Cachan <jvacher@ens-cachan.fr>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
* @file ps_lib.c
* @brief Source code for Heeger & Bergen texture synthesis.
*
* @version ?
* @author Thibaud BRIAND & Jonathan VACHER ; <thibaud.briand@ens-cachan.fr> ; <jvacher@ens-cachan.fr>
*
*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <fftw3.h>
#include <iostream>
#include "filters.h"
#include "periodic_component.h"
#include "parameters.h"
#include "bilinear_zoom.h"
#include "toolbox.h"
#include "mt19937ar.h"
#include "Eigen/Dense"

using namespace Eigen;
using namespace std;

/* M_PI is a POSIX definition */
#ifndef M_PI
/** macro definition for Pi */
#define M_PI 3.14159265358979323846
#endif                          /* !M_PI */

/*
* number of threads to use for libfftw
* (uncomment to enable parallel FFT multi-threading)
*/
#define FFTW_NTHREADS 4

 /**
 * @brief Make the analysis of a 2D array.
 *
 * The output is a list of 2+N_pyr*N_steer arrays.
 * The computation takes place in the Fourier domain.
 *
 * @image html analysis.png Example of analysis (2 scales, 4 steers)
 *
 * @param data_in input array (that you want to analyse)
 * @param list_filters, list_downsample_filters filters used during the analysis and the downsampling
 * @param size_list, size_list_downsample list of filters sizes used during the analysis and the downsampling
 * @param nx,ny array size
 * @param N_steer number of orientation bands
 * @param N_pyr number of scales
 * @return list_out the list of oriented bandpass images
 */
static void analysisPs(fftwf_complex **list_out, float **list_out_low, float *data_in, float **list_filters,
		       float **list_downsample_filters ,size_t *size_list[2],
		       size_t *size_list_downsample[2],size_t nx,size_t ny ,
		       int N_steer, int N_pyr,int option)
{
   /* allocate memory */
      size_t sx=0;
      size_t sy=0;
      size_t tx=0;
      size_t ty=0;
      fftwf_complex *fft_data_in = (fftwf_complex *) fftwf_malloc( nx*ny * sizeof(fftwf_complex));
      fftwf_complex *fft_data_in_high = (fftwf_complex *) fftwf_malloc( nx*ny * sizeof(fftwf_complex));
      fftwf_complex *fft_data_in_loop = (fftwf_complex *) fftwf_malloc( nx*ny * sizeof(fftwf_complex));
      fftwf_complex *fft_data_in_loop_loop = (fftwf_complex *) fftwf_malloc( nx*ny* sizeof(fftwf_complex));
      fftwf_complex *fft_data_in_loop_downsampled = (fftwf_complex *) fftwf_malloc(nx*ny*sizeof( fftwf_complex));
      float *data_in_noised = (float*) malloc(nx*ny*sizeof(float));

    /* Start threaded fftw if FFTW_NTHREADS is defined */
      #ifdef FFTW_NTHREADS
      //fftwf_init_threads();
      //fftwf_plan_with_nthreads(FFTW_NTHREADS);
      #endif

      if(option == 1)
      {
    /* add a small noise to avoid artefacts in the case of artificial texture */
      for(unsigned int i=0;i<nx*ny;i++) data_in_noised[i] = data_in[i]+ mt_genrand_res53()*1e-3;
      }
      else
      {
	for(unsigned int i=0;i<nx*ny;i++) data_in_noised[i] = data_in[i];
      }

    /* Compute the dft of the input */
      do_fft_real(fft_data_in,data_in_noised ,nx ,ny );

    /* high component */
      pointwise_complexfloat_multiplication( fft_data_in_high , fft_data_in , list_filters[0] , nx*ny );
      do_ifft(fft_data_in_high, list_out[0] , nx , ny );

    /* first low component */
      pointwise_complexfloat_multiplication( fft_data_in_loop , fft_data_in , list_downsample_filters[0] , nx*ny );
      do_ifft_real(fft_data_in_loop, list_out_low[0], nx, ny );

    /* recursive loop  to compute the oriented bandpass images*/
      for(int i=0;i<N_pyr;i++)
      {
		for(int j=0;j<N_steer;j++)
		{
			/* image size */
			sx=size_list[1+i*N_steer+j][0];
			sy=size_list[1+i*N_steer+j][1];

			/* apply the filters */
			pointwise_complexfloat_multiplication(fft_data_in_loop_loop,fft_data_in_loop, list_filters[1+i*N_steer+j], sx*sy);

			/* iDFT */
			do_ifft( fft_data_in_loop_loop , list_out[1+i*N_steer+j] , sx , sy );
		}

	/* DFT for the next level */
	   tx=size_list_downsample[i+1][0];
	   ty=size_list_downsample[i+1][1];
	   pointwise_complexfloat_multiplication(fft_data_in_loop,fft_data_in_loop, list_downsample_filters[1+i], tx*ty);

	/* downsample */
	      for(unsigned int l=0;l< (unsigned int) tx/2;l++)
	      {
			for(unsigned int r=0;r<(unsigned int) ty/2;r++)
			{
				for(int d=0;d<2;d++)
				{
					if((l<tx/4) && (r<ty/4)) fft_data_in_loop_downsampled[l+r*tx/2][d] = fft_data_in_loop[l+r*tx][d]/4;
					else if((l<tx/4) && (r>ty/4-1)) fft_data_in_loop_downsampled[l+r*tx/2][d] = fft_data_in_loop[l+(r+ty/2)*tx][d]/4;
					else if((l>tx/4-1) && (r<ty/4)) fft_data_in_loop_downsampled[l+r*tx/2][d] = fft_data_in_loop[l+tx/2+r*tx][d]/4;
					else fft_data_in_loop_downsampled[l+r*tx/2][d] = fft_data_in_loop[l+tx/2+(r+ty/2)*tx][d]/4;
				}
			}
	      }

	      for(unsigned int k=0;k<tx*ty/4;k++)
	      {
			for(int d=0;d<2;d++)  fft_data_in_loop[k][d] = fft_data_in_loop_downsampled[k][d];
	      }

	      /* ifft of the low band (downsampled except for the last step) */
              do_ifft_real( fft_data_in_loop , list_out_low[i+1] , tx/2 , ty/2 );
      }

    /* Compute the last low component */
       do_ifft( fft_data_in_loop , list_out[1+N_pyr*N_steer] , tx/2 , ty/2 );

    /* free memory */
       free(data_in_noised);
       fftwf_free(fft_data_in_high);
       fftwf_free(fft_data_in_loop);
       fftwf_free(fft_data_in);
       fftwf_free(fft_data_in_loop_loop);
       fftwf_free(fft_data_in_loop_downsampled);
}


 /**
 * @brief Make the synthesis of a list of oriented bandpass images.
 *
 * The output is a 2D real array.
 * The computation takes place in the Fourier domain.
 *
 * @param list_in input array (that you want to analyse)
 * @param list_filters, list_upsample_filters filters used during the synthesis and the upsampling
 * @param size_list, size_list_upsample list of filters sizes used during the synthesis and the upsampling
 * @param nx,ny array size
 * @param N_steer number of orientation bands
 * @param N_pyr number of scales
 * @return data_out ouput 2D real array
 */
static void synthesisPs( float *data_out,fftwf_complex **list_in, float **list_in_low ,float **list_filters,
			 float **list_upsample_filters ,float **list_mask_filters, size_t *size_list[2],
			 size_t *size_list_upsample[2], float *marg_stats, const MatrixXd& autocor_low, const MatrixXd& autocor_mod,
			 const MatrixXf& cross_cor_mod, const MatrixXf& cross_cor_phases, size_t nx, size_t ny, int N_steer, int N_pyr)
{
/* allocate memory */
   size_t sx,sy,tx,ty; /*number of row and columns*/
   fftwf_complex *fft_low_down = (fftwf_complex *) fftwf_malloc( nx*ny * sizeof(fftwf_complex));
   fftwf_complex *fft_out = (fftwf_complex *) fftwf_malloc( nx*ny * sizeof(fftwf_complex));
   fftwf_complex *fft_list_in = (fftwf_complex *) fftwf_malloc(nx*ny*sizeof(fftwf_complex));
   fftwf_complex *fft_low_up =  (fftwf_complex *) fftwf_malloc(nx*ny*sizeof(fftwf_complex));
   fftwf_complex *fft_steer   = (fftwf_complex *) fftwf_malloc(nx*ny*sizeof(fftwf_complex));
   fftwf_complex *fft_high =  (fftwf_complex *) fftwf_malloc(nx*ny*sizeof(fftwf_complex));

   float *list_modulus_or_Re[N_steer];
   float *list_modulus_sons[N_steer];
   float *list_angle[N_steer];
   float *list_sons_Re_Im[2*N_steer];
   float *list_modulus_or_Re_prov[1];
   list_modulus_or_Re_prov[0] = (float *) malloc(nx*ny*sizeof(float));

   for(int i=0; i<N_steer; i++)
   {
     list_modulus_or_Re[i] = (float *) malloc(nx*ny*sizeof(float));
     list_modulus_sons[i] = (float *) malloc(nx*ny*sizeof(float));
     list_angle[i]   = (float *) malloc(nx*ny*sizeof(float));
     list_sons_Re_Im[i] = (float *) malloc(nx*ny*sizeof(float));
     list_sons_Re_Im[N_steer+i] = (float *) malloc(nx*ny*sizeof(float));
   }
   float *lowpass = (float*) malloc(nx*ny*sizeof(float));
   float *highpass = (float*) malloc(nx*ny*sizeof(float));

   float m,meanmod,me,re,im;
   double v;
   
   /* Start threaded fftw if FFTW_NTHREADS is defined */
   #ifdef FFTW_NTHREADS
   //fftwf_init_threads();
   //fftwf_plan_with_nthreads(FFTW_NTHREADS);
   #endif

   /* size of central auto correlation sample */
   int Na=autocor_low.rows();
   /* to know if mod_cross_scale_cor has an effect */
   int mod[N_steer];
   for(int j=0;j<N_steer;j++) mod[j]=0;
   /* reference variance */
   float var0 = marg_stats[3];

   /* Matching of the first lowband */
      /* lowband sizes */
      sx=size_list[1+N_pyr*N_steer][0];
      sy=size_list[1+N_pyr*N_steer][1];

      for(unsigned int ind=0; ind<sx*sy ;ind++) lowpass[ind]=list_in_low[N_pyr][ind];

      /* substract the mean */
       m = mean(lowpass, sx*sy);
       for(unsigned int ind = 0; ind<sx*sy ;ind++) lowpass[ind] -= m;

      /* start adjustment */
      float vari = autocor_low((Na+1)/2-1, N_pyr*Na+(Na+1)/2-1);
      if(vari>var0*1e-4)
      {
      /* Auto correlation adjustment */
           mod_auto_cor(lowpass, autocor_low.block(0,N_pyr*Na,Na,Na), sx, sy);
      /* Skewness adjustment */
          mod_skewness(lowpass, marg_stats[N_pyr+7], sx*sy);
       /* Kurtosis adjustment */
          mod_kurtosis(lowpass, marg_stats[2*N_pyr+8], sx*sy);
      }
      else
      {
       mod_mean_variance(lowpass, 0 , vari, sx*sy);
      }
  /* Compute the DFT of the modified lowband*/
  do_fft_real( fft_low_down , lowpass , sx , sy );

  /* Loop for inversion */
  for(int i=0;i<N_pyr;i++)
  {
    /* define size */
     tx=size_list_upsample[N_pyr-i][0];
     ty=size_list_upsample[N_pyr-i][1];

    /* Upsampling */
    for(unsigned int r=0;r<(unsigned int) ty;r++)
    {
      for(unsigned int l=0;l<(unsigned int) tx;l++)
      {
	for(int d=0;d<2;d++)
	{
		if ((l<tx/4) && (r<ty/4))
		fft_low_up[l+r*tx][d] = fft_low_down[l+r*tx/2][d]*4;
		else if((l<tx/4) && (r>3*ty/4-1))
		fft_low_up[l+r*tx][d] = fft_low_down[l+(r-ty/2)*tx/2][d]*4;
		else if((l>3*tx/4-1) && (r<ty/4))
		fft_low_up[l+r*tx][d] = fft_low_down[l-tx/2+r*tx/2][d]*4;
		else if((l>3*tx/4-1) && (r>3*ty/4-1))
		fft_low_up[l+r*tx][d] = fft_low_down[l-tx/2+(r-ty/2)*tx/2][d]*4;
		else
		fft_low_up[l+r*tx][d]=0;
	}
      }
    }

    /* First computation: lowpass */
    pointwise_complexfloat_multiplication(fft_low_up, fft_low_up, list_upsample_filters[N_pyr-i], tx*ty );
    /* compute the modulus of the subbands */
    for(int j=0; j<N_steer; j++)
    {
      for(unsigned int ind = 0; ind<tx*ty ; ind ++)
      {
	re = list_in[(N_pyr-i)*N_steer-j][ind][0];
	im = list_in[(N_pyr-i)*N_steer-j][ind][1];
	list_modulus_or_Re[j][ind] = sqrt(re*re+ im*im);
	if(im==0 && re==0) list_angle[j][ind]=0;
	else
        list_angle[j][ind]=atan2(im,re);
      }

      /* substract mean */
      m = mean(list_modulus_or_Re[j], tx*ty);
       for(unsigned int ind = 0; ind<tx*ty ; ind ++) list_modulus_or_Re[j][ind] -= m;
    }

    /* crossscale correlation des magnitudes */
    if(i==0)
    {
       mod_cross_cor(list_modulus_or_Re, cross_cor_mod.block(0, (N_pyr-1-i)*N_steer, N_steer, N_steer), tx*ty);
      for(int j=0; j<N_steer; j++)
      {
        mod_auto_cor(list_modulus_or_Re[j], autocor_mod.block(0,(j+N_steer*(N_pyr-i-1))*Na,Na,Na), tx, ty);
        meanmod = marg_stats[9+2*N_pyr+(N_pyr-i-1)*N_steer+j];
        for(unsigned int ind = 0; ind<tx*ty; ind++)
        {
	  list_modulus_or_Re[j][ind] += meanmod;
	  list_in[(N_pyr-i)*N_steer-j][ind][0]  = (list_modulus_or_Re[j][ind]>=0)*list_modulus_or_Re[j][ind]*cos(list_angle[j][ind]);
	  list_in[(N_pyr-i)*N_steer-j][ind][1]  = (list_modulus_or_Re[j][ind]>=0)*list_modulus_or_Re[j][ind]*sin(list_angle[j][ind]);
	  list_modulus_or_Re[j][ind] = list_in[(N_pyr-i)*N_steer-j][ind][0];
        }
      }
    }
    else
    {
      for(int j=0; j<N_steer;j++)
      {
	do_fft( fft_list_in , list_in[(N_pyr-i+1)*N_steer-j] , tx/2 , ty/2 );
	/* Upsampling of the sons*/
	for(unsigned int r=0;r<(unsigned int) ty;r++)
	{
	  for(unsigned int l=0;l<(unsigned int) tx;l++)
	  {
	    for(int d=0;d<2;d++)
	    {
		  if ((l<tx/4) && (r<ty/4))
		  fft_steer[l+r*tx][d] = fft_list_in[l+r*tx/2][d]*4;
		  else if((l<tx/4) && (r>3*ty/4-1))
		  fft_steer[l+r*tx][d] = fft_list_in[l+(r-ty/2)*tx/2][d]*4;
		  else if((l>3*tx/4-1) && (r<ty/4))
		  fft_steer[l+r*tx][d] = fft_list_in[l-tx/2+r*tx/2][d]*4;
		  else if((l>3*tx/4-1) && (r>3*ty/4-1))
		  fft_steer[l+r*tx][d] = fft_list_in[l-tx/2+(r-ty/2)*tx/2][d]*4;
		  else
		  fft_steer[l+r*tx][d]=0;
	    }
	  }
	}
	
	do_ifft(fft_steer, fft_out, tx, ty);
	
	/* keep separately modulus, real and imaginary parts of the sons */
	for(unsigned int ind=0;ind<tx*ty;ind++)
	{
	  re = fft_out[ind][0];
	  im = fft_out[ind][1];
	  list_modulus_sons[j][ind]=sqrt(re*re+im*im);
	  list_sons_Re_Im[j][ind]=(re*re-im*im)/list_modulus_sons[j][ind];
	  list_sons_Re_Im[N_steer+j][ind]=2*im*re/list_modulus_sons[j][ind];
	}
	
	me=mean(list_modulus_sons[j], tx*ty);
	
	for(unsigned int ind=0;ind<tx*ty;ind++)
	{
	  list_modulus_sons[j][ind]-=me;
	}
	
      }

      /* adjust cross correlation of modulus of sub bands */
         mod_cross_scale_cor(list_modulus_or_Re,list_modulus_sons, cross_cor_mod.block(0, (N_pyr-1-i)*N_steer, N_steer, N_steer),
 			  cross_cor_mod.block(0, (2*N_pyr-1-i)*N_steer, N_steer, N_steer), tx*ty);

      /* adjust auto correlation of modulus of each sub band  and impose their mean */
      for(int j=0; j<N_steer; j++)
      {
        mod_auto_cor(list_modulus_or_Re[j], autocor_mod.block(0,(j+N_steer*(N_pyr-i-1))*Na,Na,Na), tx, ty);
        meanmod = marg_stats[9+2*N_pyr+(N_pyr-i-1)*N_steer+j];
        for(unsigned int ind = 0; ind<tx*ty; ind++)
        {
 	  list_modulus_or_Re[j][ind] += meanmod;
	  list_modulus_or_Re[j][ind] = (list_modulus_or_Re[j][ind]>=0)*list_modulus_or_Re[j][ind];
	  list_modulus_or_Re_prov[0][ind] = list_modulus_or_Re[j][ind]*cos(list_angle[j][ind]);
	  list_in[(N_pyr-i)*N_steer-j][ind][1] = list_modulus_or_Re[j][ind]*sin(list_angle[j][ind]);
        }
        /* cross correlation of real parts with both real and imaginary part of sons */
        v = compute_moment(list_modulus_or_Re_prov[0], 2, tx*ty);
	MatrixXf V(1,1);
	V << v;
	mod[j]=mod_cross_scale_cor(list_modulus_or_Re_prov, list_sons_Re_Im, V, cross_cor_phases.block(j, 2*(N_pyr-1-i)*N_steer, 1, 2*N_steer), tx*ty);
	if(mod[j]==1)
	{
	 for(unsigned int ind = 0; ind<tx*ty; ind++)
         {
	    list_modulus_or_Re[j][ind] = list_modulus_or_Re_prov[0][ind];
         }
	}
	else
	{
	 for(unsigned int ind = 0; ind<tx*ty; ind++)
         {
	   list_in[(N_pyr-i)*N_steer-j][ind][0] = list_modulus_or_Re_prov[0][ind];
         }
	}
      }
    }

    /* mask to get an analytic signal */
    for(int j=0;j<N_steer;j++)
    {
      if(mod[j]==1 || i==0)
      {
        do_fft_real(fft_steer, list_modulus_or_Re[j], tx, ty);
        pointwise_complexfloat_multiplication(fft_steer, fft_steer,
				            list_mask_filters[N_steer*(N_pyr-i)-j-1], tx*ty);
        do_ifft(fft_steer,list_in[(N_pyr-i)*N_steer-j],tx,ty);
        for(unsigned int ind=0;ind<tx*ty;ind++) list_modulus_or_Re[j][ind]=list_in[(N_pyr-i)*N_steer-j][ind][0];
      }
      else
      {
        do_fft(fft_steer, list_in[(N_pyr-i)*N_steer-j], tx, ty);
        pointwise_complexfloat_multiplication(fft_steer, fft_steer,
				            list_mask_filters[N_steer*(N_pyr-i)-j-1], tx*ty);
        do_ifft(fft_steer,list_in[(N_pyr-i)*N_steer-j],tx,ty);
        for(unsigned int ind=0;ind<tx*ty;ind++) list_modulus_or_Re[j][ind]=list_in[(N_pyr-i)*N_steer-j][ind][0];
      }
    }

    /* Computation for steerpass */
    for(int j=0;j<N_steer;j++)
    {
      /* apply the filters and add */
      do_fft_real(fft_list_in, list_modulus_or_Re[j], tx, ty);
      pointwise_complexfloat_multiplication(fft_steer, fft_list_in,
					    list_filters[(N_pyr-i)*N_steer-j], tx*ty);
      for(unsigned int k=0;k<tx*ty;k++)
      {
	for(int d=0;d<2;d++)
		    fft_low_up[k][d]+=fft_steer[k][d];
      }
    }

    /* the low-band has to be real */
    do_ifft_real(fft_low_up, lowpass, tx, ty);

    vari = autocor_low((Na+1)/2-1,(N_pyr-1-i)*Na+(Na+1)/2-1);
     if(vari>var0*1e-4)
     {
      /* Auto correlation adjustment */
           mod_auto_cor(lowpass, autocor_low.block(0,(N_pyr-i-1)*Na,Na,Na), tx, ty);
      /* Skewness adjustment */
           mod_skewness(lowpass, marg_stats[7+N_pyr-1-i], tx*ty);
       /* Kurtosis adjustment */
           mod_kurtosis(lowpass, marg_stats[8+2*N_pyr-1-i], tx*ty);
      }
      else
      {
       mod_mean_variance(lowpass, 0 , vari, tx*ty);
      }

      /* update fft_low_down */
      do_fft_real(fft_low_down,lowpass, tx, ty);
  }

  /* last step : high pass and low pass */
       /* low pass */
       pointwise_complexfloat_multiplication(fft_low_down, fft_low_down,
					     list_upsample_filters[0], nx*ny);

       /* high pass */
          /* variance Adjustment */
	  for(unsigned int ind =0; ind<nx*ny;ind++) highpass[ind] = list_in[0][ind][0];
	  vari = compute_moment(highpass, 2, nx*ny);
 	  if(vari>marg_stats[6]) mod_mean_variance(highpass, 0 , marg_stats[6], nx*ny);

	  /* Highpass filter multiplication */
	  do_fft_real(fft_list_in, highpass, nx, ny);
          pointwise_complexfloat_multiplication(fft_high, fft_list_in, list_filters[0], nx*ny);

  for(unsigned int k=0;k<nx*ny;k++) {for(int d=0;d<2;d++) fft_out[k][d]=fft_low_down[k][d]+fft_high[k][d];}

  /* final iDFT */
  do_ifft_real(fft_out, data_out, nx, ny);

  /* Marginal statistics adjustment */

    /* skewness adjustment */
      mod_skewness(data_out, marg_stats[4], nx*ny);
     /* kurtosis adjustment */
      mod_kurtosis(data_out, marg_stats[5], nx*ny);
    /* Mean and variance adjustment */
      mod_mean_variance(data_out, marg_stats[2] , marg_stats[3], nx*ny);
     /* range adjustment */
      mod_range(data_out, marg_stats[1], marg_stats[0], nx*ny);


  /* free memory */
    fftwf_free(fft_out);
    fftwf_free(fft_high);
    fftwf_free(fft_low_down);
    fftwf_free(fft_list_in);
    fftwf_free(fft_low_up);
    fftwf_free(fft_steer);
    for(int i=0; i<N_steer; i++)
    {
      free(list_angle[i]);
      free(list_modulus_sons[i]);
      free(list_modulus_or_Re[i]);
      free(list_sons_Re_Im[i]);
      free(list_sons_Re_Im[N_steer+i]);
    }
    free(list_modulus_or_Re_prov[0]);
    free(lowpass);
    free(highpass);
}
/* Texture synthesis section */

/**
 * @brief Compute the set of constraints on a pyramid.
 *
 * @param tab input array
 * @param N size of the array
 * @return max of tab
 */
template <typename Derivedf,typename Derivedd>
void compute_constraints(float *marg_stats, MatrixBase<Derivedd> const & autocor_low, MatrixBase<Derivedd> const & autocor_mod, MatrixBase<Derivedf> const & cross_cor_mod,
			 MatrixBase<Derivedf> const & cross_cor_phases, float *data_in, fftwf_complex **list_in, float **list_low_in,
			 int N_steer, int N_pyr, int Na, size_t nx, size_t ny)
{
  MatrixXd autocor_low_in(Na, Na*(N_pyr+1));
  MatrixXd autocor_mod_in(Na, Na*N_steer*N_pyr);
  MatrixXf cross_cor_mod_in(N_steer, (2*N_pyr-1)*N_steer);
  MatrixXf cross_cor_phases_in(N_steer, 2*(N_pyr-1)*N_steer);

 float re,im;
 int N=nx*ny;
 int s,tx,ty;

 /* compute marginal statistics of the image */
 marg_stats[0] = (float) max(data_in,N);
 marg_stats[1] = (float) min(data_in,N);
 marg_stats[2] = (float) mean(data_in,N);
 marg_stats[3] = (float) compute_moment(data_in,2,N);
 marg_stats[4] = (float) compute_skewness(data_in,N);
 marg_stats[5] = (float) compute_kurtosis(data_in,N);

 /* compute variance of high pass */
 float *high = (float*) malloc(N*sizeof(float));
 for(int i=0;i<N;i++) high[i]=list_in[0][i][0];
 marg_stats[6] = (float) compute_moment(high,2,N);
 free(high);

  // compute marginal statistics of low passes
 for(int ii=0;ii<(N_pyr+1);ii++)
 {
   s=(1<<ii)*(1<<ii);
   marg_stats[7+ii] = (float) compute_skewness(list_low_in[ii],(int) N/s);
   marg_stats[8+N_pyr+ii] = (float) compute_kurtosis(list_low_in[ii],(int) N/s);
 }

   // compute auto correlation of low passes
 for(int i=0;i<(N_pyr+1);i++)
 {
  s=(1<<i);
  autocor_low_in.block(0,i*Na,Na,Na) = compute_auto_cor(list_low_in[i],(size_t) nx/s, (size_t) ny/s, Na);
 }

 float *mod[N_steer];
 float *Real[N_steer];
 float *mod_prev[N_steer];
 float *Real_Imag_prev[2*N_steer];
 fftwf_complex *list_in_fft = (fftwf_complex*) fftwf_malloc(N_steer*N*sizeof(fftwf_complex));
 fftwf_complex *list_prev_fft = (fftwf_complex*) fftwf_malloc(N_steer*N*sizeof(fftwf_complex));
 fftwf_complex *list_prev = (fftwf_complex*) fftwf_malloc(N_steer*N*sizeof(fftwf_complex));

 for(int ll=0;ll<N_steer;ll++)
  {
    mod[ll] = (float*) malloc(N*sizeof(float*));
    mod_prev[ll] = (float*) malloc(N*sizeof(float));
    Real[ll] = (float*) malloc(N*sizeof(float));
    Real_Imag_prev[ll] = (float*) malloc(N*sizeof(float));
    Real_Imag_prev[N_steer+ll] = (float*) malloc(N*sizeof(float));
  }

 for(int k=0;k<N_pyr;k++)
 {
   s=(1<<k);
   tx=nx/s;
   ty=ny/s;

   for(int l=0;l<N_steer;l++)
   {
     for(int iii=0;iii<tx*ty;iii++)
     {
       re = list_in[N_steer*(k+1)-l][iii][0];
       im = list_in[N_steer*(k+1)-l][iii][1];
      Real[l][iii] = re;
      mod[l][iii] = sqrt(re*re+im*im);
     }
      /* compute mean of sub bands modulus and their auto correlation */
      marg_stats[9+2*N_pyr+l+k*N_steer]=mean(mod[l], tx*ty);
      autocor_mod_in.block(0,(l+k*N_steer)*Na,Na,Na) = compute_auto_cor(mod[l], tx, ty, Na);

    if(k<N_pyr-1)
    {
     // up sample the coarser scale
     do_fft(list_in_fft, list_in[N_steer*(k+2)-l], tx/2 , ty/2);

     for( int y=0;y< ty;y++)
     {
       for( int x=0;x< tx;x++)
       {
         for(int d=0;d<2;d++)
         {
	   if ((x<tx/4) && (y<ty/4))
	     list_prev_fft[x+y*tx][d] = list_in_fft[x+y*tx/2][d]*4;
	   else if((x<tx/4) && (y>3*ty/4-1))
	     list_prev_fft[x+y*tx][d] = list_in_fft[x+(y-ty/2)*tx/2][d]*4;
	   else if((x>3*tx/4-1) && (y<ty/4))
	     list_prev_fft[x+y*tx][d] = list_in_fft[x-tx/2+y*tx/2][d]*4;
	   else if((x>3*tx/4-1) && (y>3*ty/4-1))
	     list_prev_fft[x+y*tx][d] = list_in_fft[x-tx/2+(y-ty/2)*tx/2][d]*4;
	   else
	     list_prev_fft[x+y*tx][d]=0;
          }
        }
      }

      do_ifft(list_prev_fft, list_prev, tx , ty);
    }

   for(int i=0;i<tx*ty;i++)
   {
     re =list_prev[i][0];
     im =list_prev[i][1];
    mod_prev[l][i] = sqrt(re*re+im*im);
    Real_Imag_prev[l][i] = (re*re-im*im)/mod_prev[l][i];
    Real_Imag_prev[N_steer+l][i] = 2*im*re/mod_prev[l][i];
   }
  }
   /* compute cross correlation and cross scale correlation of modulus
      then the cross scale phases statistics
   */

  /* effective computation of modulus cross scale correlation */
   cross_cor_mod_in.block(0, k*N_steer, N_steer, N_steer) = compute_cross_cor(mod, N_steer, tx*ty);

  if(k<N_pyr-1)
  {
    cross_cor_phases_in.block(0, 2*k*N_steer, N_steer, 2*N_steer) = compute_cross_scale_cor(Real, Real_Imag_prev, N_steer, 2*N_steer, tx*ty,0);
    cross_cor_mod_in.block(0, (k+N_pyr)*N_steer, N_steer, N_steer) = compute_cross_scale_cor(mod, mod_prev, N_steer, N_steer, tx*ty,1);
  }
 }

 for(int f=0;f<N_steer;f++)
 {
    free(mod[f]);
    free(mod_prev[f]); 
    free(Real[f]);
    free(Real_Imag_prev[f]);
    free(Real_Imag_prev[N_steer+f]);
 }

 fftwf_free(list_in_fft);
 fftwf_free(list_prev_fft);
 fftwf_free(list_prev);

  const_cast< MatrixBase<Derivedd> & >(autocor_low)=autocor_low_in;
  const_cast< MatrixBase<Derivedd> & >(autocor_mod)=autocor_mod_in;
  const_cast< MatrixBase<Derivedf> & >(cross_cor_mod)=cross_cor_mod_in;
  const_cast< MatrixBase<Derivedf> & >(cross_cor_phases)=cross_cor_phases_in;
}

 /**
 * @brief Compute a texture of gray level according to the algorithm of Heeger
                 & Bergen (but using the periodic component).
 *
* @param data_in input array (sample texture)
* @param nxout,nyout output size
* @param nxin,nyin input size
* @param N_steer number of orientation bands
* @param N_pyr number of scales
* @param N_iteration number of iterations
* @param noise initial image
* @param list_filters_texture, list_downsample_filters_texture_filters filters
* used during the synthesis and the upsampling (texture)
* @param list_filters_sample, list_downsample_filters_sample filters used
* during the synthesis and the upsampling (sample)
* @param size_list_texture,size_list_downsample_texture list of sizes
* @param size_list_sample, size_list_downsample_sample list of sizes
* @param edge_handling option for edge handling
* @return data_out output array (synthesized texture)
 */
static void createTexture(float *data_out, float *data_in,  size_t nxout, size_t nyout, size_t nxin, size_t nyin,
			  int N_steer, int N_pyr, int N_iteration, int Na, float *noise, int factor_noise,
			  float *list_filters_sample[],float **list_downsample_filters_sample,
			  size_t *size_list_sample[2],size_t *size_list_downsample_sample[2],
			  float **list_filters_texture_analysis, float **list_downsample_filters_texture,
			  float **list_filters_texture_synthesis, float **list_mask_filters,
			  size_t *size_list_texture[2], size_t *size_list_downsample_texture[2], int edge_handling)
{
  /* memory allocation */

      /* list of constraint */
      float *marg_stats = (float *) malloc((9+N_pyr*(N_steer+2))*sizeof(float));
      float *data_old = (float *) malloc(nxout*nyout*sizeof(float));
      MatrixXd autocor_low(Na, Na*(N_pyr+1));
      MatrixXd autocor_mod(Na, Na*N_steer*N_pyr);
      MatrixXf cross_cor_mod(N_steer, (2*N_pyr-1)*N_steer);
      MatrixXf cross_cor_phases(N_steer, 2*(N_pyr-1)*N_steer);

      /* for the analysis : set of images */
      fftwf_complex *list_sample[2+N_pyr*N_steer];
      float *list_sample_low[1+N_pyr];
      fftwf_complex *list_texture[2+N_pyr*N_steer];
      float *list_texture_low[1+N_pyr];
      fftwf_complex *list_sample_mirror[2+N_pyr*N_steer];
      float *list_sample_mirror_low[1+N_pyr];

      /* sizes */
      size_t *size_list_final_vector = (size_t *) malloc(2*(2+N_pyr*N_steer)*sizeof(size_t));
      size_t *size_list_final[(2+N_pyr*N_steer)];
		    for(int k=0;k<2+N_pyr*N_steer;k++)
		      size_list_final[k]=size_list_final_vector+2*k;

      /* periodic_component case */
	if(edge_handling == 0 || edge_handling == 2)
	{
	  for(int k=0;k<(2+N_pyr*N_steer);k++)
	  {
	    size_list_final[k][0]=size_list_sample[k][0];
	    size_list_final[k][1]=size_list_sample[k][1];
	  }
	}
      /* mirror symmetrization case */
	else
	{
	  for(int k=0;k<(2+N_pyr*N_steer);k++)
	  {
	    size_list_final[k][0]=size_list_sample[k][0]/2;
	    size_list_final[k][1]=size_list_sample[k][1]/2;
	  }
	}

      /* allocate memory according to the size of the images*/
	if(edge_handling == 0 || edge_handling == 2)
	{
	 for(int i=0; i<2+N_pyr*N_steer;i++)
	    list_sample[i] = (fftwf_complex *) 
	    fftwf_malloc(size_list_sample[i][0]*size_list_sample[i][1]*sizeof(fftwf_complex));
	 for(int i=0; i<1+N_pyr;i++)
	    list_sample_low[i] = (float *)
	    malloc(size_list_downsample_sample[i][0]*size_list_downsample_sample[i][1]*sizeof(float));
	}
	else if(edge_handling==1)
	{
	    for(int i=0; i<2+N_pyr*N_steer;i++)
	    {
	     list_sample_mirror[i]= (fftwf_complex *) 
	     fftwf_malloc(size_list_sample[i][0]*size_list_sample[i][1]*sizeof(fftwf_complex));

	     for(unsigned int m=0;m<size_list_sample[i][0]*size_list_sample[i][1];m++){
	       list_sample_mirror[i][m][0]=0;
	       list_sample_mirror[i][m][1]=0;
	      }

	     list_sample[i] = (fftwf_complex *) 
	     fftwf_malloc(size_list_final[i][0]*size_list_final[i][1]*sizeof(fftwf_complex));
	    }

	    for(int i=0; i<1+N_pyr;i++)
	    {
	      list_sample_low[i] = (float *) malloc(size_list_downsample_sample[i][0]*size_list_downsample_sample[i][1]/4*sizeof(float));
              list_sample_mirror_low[i] = (float *) malloc(size_list_downsample_sample[i][0]*size_list_downsample_sample[i][1]*sizeof(float));
	      for(unsigned int m=0;m<size_list_downsample_sample[i][0]*size_list_downsample_sample[i][1];m++)
	      {
	       list_sample_mirror_low[i][m]=0;
	      }
	    }
	}

	/* for the synthesized texture */
	      /* allocate memory */
		  for(int i=0; i<2+N_pyr*N_steer;i++)
		    list_texture[i] = (fftwf_complex *) fftwf_malloc(size_list_texture[i][0]*size_list_texture[i][1]*sizeof(fftwf_complex));
		  for(int i=0; i<1+N_pyr;i++)
	            list_texture_low[i] = (float *) malloc(size_list_downsample_texture[i][0]*size_list_downsample_texture[i][1]*sizeof(float));

     /* Analysis of the sample */
      if(edge_handling == 0 || edge_handling == 2)
      {
      analysisPs(list_sample, list_sample_low, data_in, list_filters_sample,
		 list_downsample_filters_sample, size_list_sample,
		 size_list_downsample_sample, nxin, nyin, N_steer, N_pyr,1);
      }
      else if(edge_handling==1)
      {
	/* compute the mirror symmetrization */
	float *data_mirror = (float*) malloc(4*nxin*nyin*sizeof(float));
	for(unsigned int i=0;i<nxin;i++)
	 {
	  for(unsigned int j=0;j<nyin;j++)
	   {
	    data_mirror[i+2*j*nxin]=data_in[i+j*nxin];
	    data_mirror[i+(2*j+1)*nxin]=data_in[nxin-1-i+j*nxin];
	    data_mirror[2*nyin*nxin+i+2*j*nxin]=data_in[i+(nyin-j-1)*nxin];
	    data_mirror[2*nyin*nxin+i+(2*j+1)*nxin]=data_in[nxin-i-1+(nyin-j-1)*nxin];
	   }
	 }

	 /* analysis */
	 analysisPs(list_sample_mirror, list_sample_mirror_low, data_mirror, list_filters_sample,
		    list_downsample_filters_sample, size_list_sample,
		    size_list_downsample_sample, 2*nxin, 2*nyin, N_steer, N_pyr,1);

	 /* keep the left-hand top corner  */
	 for(int d=0;d<2+N_pyr*N_steer;d++)
	 {
	   for(unsigned int i=0;i<size_list_final[d][0];i++)
	   {
	    for(unsigned int j=0;j<size_list_final[d][1];j++){
	      list_sample[d][i+j*size_list_final[d][0]][0] = list_sample_mirror[d][i+j*size_list_sample[d][0]][0];
	      list_sample[d][i+j*size_list_final[d][0]][1] = list_sample_mirror[d][i+j*size_list_sample[d][0]][1];
	    }
	   }
	 }
	 for(int d=0;d<1+N_pyr;d++){
	   for(unsigned int i=0;i<size_list_downsample_sample[d][0]/2;i++)
	   {
	    for(unsigned j=0;j<size_list_downsample_sample[d][1]/2;j++)
	    {
	      list_sample_low[d][i+j*size_list_downsample_sample[d][0]/2] = list_sample_mirror_low[d][i+j*size_list_downsample_sample[d][0]];
	    }
	   }
	 }

	/* free */
	free(data_mirror);
	for(int i=0;i<2+N_pyr*N_steer;i++) fftwf_free(list_sample_mirror[i]);
	for(int i=0;i<1+N_pyr;i++) free(list_sample_mirror_low[i]);
      }

      /* Compute the list of constraint */
      compute_constraints(marg_stats, autocor_low, autocor_mod, cross_cor_mod, cross_cor_phases, data_in, list_sample, list_sample_low, N_steer, N_pyr, Na, nxin, nyin);
      
      /* initialization*/
      for(unsigned int i=0;i<nxout*nyout;i++) {
	data_out[i]=noise[i];
	data_old[i]=noise[i];
      }

        if(factor_noise == 0)
	{
	  mod_mean_variance(data_out, marg_stats[2] , marg_stats[3], nxout*nyout);
	}

      /* Loop in order to create the texture */
      for(int k=0;k<N_iteration;k++)
      {
	/* Create the oriented bandpass images for the texture */
	  analysisPs(list_texture, list_texture_low, data_out, list_filters_texture_analysis,
		     list_downsample_filters_texture, size_list_texture,
	      size_list_downsample_texture, nxout, nyout, N_steer, N_pyr,0);

	/* Compute the texture : synthesis */
	  synthesisPs(data_out,list_texture, list_texture_low, list_filters_texture_synthesis,
		      list_downsample_filters_texture, list_mask_filters, size_list_texture,
	       size_list_downsample_texture, marg_stats, autocor_low, autocor_mod, cross_cor_mod, cross_cor_phases, nxout, nyout, N_steer, N_pyr);

	  if(k<N_iteration-1)
	  {
	    for(unsigned int i=0;i<nxout*nyout;i++) {
	      data_out[i] += 0.8*(data_out[i]-data_old[i]);
	      data_old[i]=data_out[i];
	    }
	  }
	}

	
  /* free memory */
      for(int i=0;i<2+N_pyr*N_steer;i++)
      {
	fftwf_free(list_sample[i]);
	fftwf_free(list_texture[i]);
      }
      for(int i=0;i<1+N_pyr;i++)
      {
	free(list_sample_low[i]);
	free(list_texture_low[i]);
      }
      free(size_list_final_vector);
      free(marg_stats);
      free(data_old);
}

/**
 * @brief Compute the synthesized texture according to the algorithm of Heeger
 *                & Bergen (but using the periodic component)
 * @param[in] data_in_rgb input color array (sample texture)
 * @param nxout,nyout output size
 * @param nxin,nyin input size
 * @param N_steer number of orientation bands
 * @param N_pyr number of scales
 * @param N_iteration number of iterations
 * @param noise initial color image
 * @param edge_handling option for edge handling
 * @param smooth option for the add of the smooth component
 * @return data_out_rgb output color array (synthesized texture)
 */
void ps(float **data_out_rgb, float **data_in_rgb, size_t nxout,
	size_t nyout, size_t nxin, size_t nyin, int N_steer, int N_pyr,
	int N_iteration, int Na, float **noise, int factor_noise, int edge_handling, int smooth,int color)
{
  /* memory allocation */
     /* 3D images */

     float *data_in_per = (float*) malloc(color * nxin * nyin * sizeof(float));
     float *data_in_rgb_per[color];
     for(int d=0;d<color;d++)
		data_in_rgb_per[d] = data_in_per + d*nxin*nyin;

     /* PCA */
     float data_in_rgb_mean[3];
     float M_eigen_vectors[3][3];

     float *data_in_pca = (float*) malloc(color * nxin * nyin * sizeof(float));
     float *data_in_rgb_pca[color];
         for(int d=0;d<color;d++)
		data_in_rgb_pca[d] = data_in_pca+d*nxin*nyin;

     float *data_out_pca = (float*) malloc(color * nxout * nyout * sizeof(float));
     float *data_out_rgb_pca[color];
         for(int d=0;d<color;d++)
		data_out_rgb_pca[d] = data_out_pca+d*nxout*nyout;

   /* Compute the periodic component of the input */
     if(edge_handling == 0)
     {
       for(int i=0; i<color ; i++ )  periodic_component(data_in_rgb_per[i],data_in_rgb[i],nxin,nyin);
     }
     else
     {
       for(int d=0;d<color;d++)
       {
	 for(unsigned int i=0; i<nxin*nyin;i++) data_in_rgb_per[d][i] = data_in_rgb[d][i];
       }
     }

     if(color ==3)
     {
    /* Make the change of basis RGB--> PCA basis */
      pca( M_eigen_vectors, data_in_rgb_mean , data_in_rgb_per, nxin , nyin );
      for(int i=0;i<3;i++)
      {
	for(unsigned int j=0;j< nxin*nyin;j++)
	{
	data_in_rgb_pca[i][j]=0;
	for(int k=0;k<3; k++)
	{
	  data_in_rgb_pca[i][j] +=
	  M_eigen_vectors[k][i]*(data_in_rgb_per[k][j]-data_in_rgb_mean[k]);
	}
	}
      }
      if(factor_noise ==1)
      {
	float noise_in_rgb_mean[3];
	float *noise_pca[3];
	for(int i=0;i<3;i++) noise_pca[i] = (float *) malloc(nxout*nyout*sizeof(float));

	for(int i=0;i<3;i++)
	{
	  noise_in_rgb_mean[i] = mean(noise[i],nxout*nyout);
	  for(unsigned int j=0;j<nxout*nyout;j++)
	  {
	  noise_pca[i][j]= (float) 0;
	  for(int k=0;k<3; k++)
	    {
	    noise_pca[i][j] += (float)
	    M_eigen_vectors[k][i]*(noise[k][j]-noise_in_rgb_mean[k]);
	    }
	  }
	 }

	for(int i=0;i<3;i++)
	{
	  for(unsigned int j=0;j<nxout*nyout;j++)
	  {
	  noise[i][j]= noise_pca[i][j];
	  }
	}

	for(int i=0;i<3;i++) free(noise_pca[i]);
       }
     }
     else
     {
        for(unsigned int j=0;j< nxin*nyin;j++) data_in_rgb_pca[0][j]=data_in_rgb_per[0][j];
     }


  /* Compute the set of filters */
    /* define the factor of reversibility alpha */
	float alpha;
	alpha = compute_alpha(N_steer);

      /* memory allocation for the filters and their sizes */
      float *list_filters_sample[2+N_pyr*N_steer];
      float *list_downsample_filters_sample[1+N_pyr];

      size_t *size_list_sample_vector = (size_t *) malloc(2*(2+N_pyr*N_steer)*sizeof(size_t));
      size_t *size_list_downsample_sample_vector = (size_t *) malloc(2*(1+N_pyr)*sizeof(size_t));

      size_t *size_list_sample[2+N_pyr*N_steer];
	    for(int k=0;k<2+N_pyr*N_steer;k++) size_list_sample[k]=size_list_sample_vector+2*k;
      size_t *size_list_downsample_sample[1+N_pyr];
		    for(int k=0;k<1+N_pyr;k++) size_list_downsample_sample[k]=size_list_downsample_sample_vector+2*k;

      float *list_filters_texture_analysis[2+N_pyr*N_steer];
      float *list_filters_texture_synthesis[2+N_pyr*N_steer];
      float *list_downsample_filters_texture[1+N_pyr];
      float *list_mask_filters[N_pyr*N_steer];

      size_t *size_list_texture_vector = (size_t *) malloc(2*(2+N_pyr*N_steer)*sizeof(size_t));
      size_t *size_list_downsample_texture_vector = (size_t *) malloc(2*(1+N_pyr)*sizeof(size_t));

      size_t *size_list_texture[2+N_pyr*N_steer];
		    for(int k=0;k<2+N_pyr*N_steer;k++) size_list_texture[k]=size_list_texture_vector+2*k;
      size_t *size_list_downsample_texture[1+N_pyr];
		    for(int k=0;k<1+N_pyr;k++) size_list_downsample_texture[k]=
		      size_list_downsample_texture_vector+2*k;

      /* Create the filters and the lists of sizes */

	  /* for the sample */
	     if(edge_handling == 0 || edge_handling == 2)
	      {
		size_filters(size_list_sample , size_list_downsample_sample ,
			     nxin , nyin , N_steer, N_pyr);
	      }
	      else
	      {
		size_filters(size_list_sample , size_list_downsample_sample ,
		             2*nxin , 2*nyin , N_steer, N_pyr);
	      }

	  /* allocate memory */
	  for(int i=0; i<2+N_pyr*N_steer;i++) 
	  {
	    list_filters_sample[i]= (float *) 
	    malloc(size_list_sample[i][0]*size_list_sample[i][1]*sizeof(float *));
	  }
	  for(int i=0; i<1+N_pyr; i++)
	  {
	      list_downsample_filters_sample[i] = (float *) malloc(size_list_downsample_sample[i][0]*size_list_downsample_sample[i][1]*sizeof(float *));
	  }

	     if(edge_handling == 0 || edge_handling == 2)
	      {
		filters_analysis(list_filters_sample, list_downsample_filters_sample , size_list_sample ,
			size_list_downsample_sample , nxin , nyin , N_steer, N_pyr, alpha);
	      }
	      else
	      {
		filters_analysis(list_filters_sample, list_downsample_filters_sample , size_list_sample ,
			size_list_downsample_sample , 2*nxin , 2*nyin , N_steer, N_pyr, alpha);
	      }

          /* for the synthesized texture */
	      size_filters(size_list_texture , size_list_downsample_texture , nxout , nyout , N_steer, N_pyr);

	      /* allocate memory */
		  for(int i=0; i<2+N_pyr*N_steer;i++)
		  {
		    list_filters_texture_analysis[i] = (float *) malloc(size_list_texture[i][0]*size_list_texture[i][1]*sizeof(float));
		    list_filters_texture_synthesis[i] = (float *) malloc(size_list_texture[i][0]*size_list_texture[i][1]*sizeof(float));
		  }
		  for(int i=0; i<1+N_pyr; i++)
		  {
		    list_downsample_filters_texture[i] = (float *)
		    malloc(size_list_downsample_texture[i][0]*size_list_downsample_texture[i][1]*sizeof(float));
		  }
		  for(int i=0; i<N_pyr; i++){
		    for(int j=0;j<N_steer; j++) {
		    list_mask_filters[i*N_steer+j] = (float *)
		    malloc(size_list_texture[1+i*N_steer][0]*size_list_texture[1+i*N_steer][1]*sizeof(float));
		    }
		  }
		  
		  filters_analysis(list_filters_texture_analysis, list_downsample_filters_texture ,
			  size_list_texture , size_list_downsample_texture ,
			  nxout , nyout , N_steer, N_pyr, alpha);
		  filters_synthesis(list_filters_texture_synthesis, list_mask_filters ,
			  size_list_texture ,
			  nxout , nyout , N_steer, N_pyr, alpha);
		
  /* Compute the texture for every color component */
  for(int i=0;i<color;i++) createTexture(data_out_rgb_pca[i], data_in_rgb_pca[i], nxout, nyout, nxin, nyin, N_steer, N_pyr,
				      N_iteration , Na, noise[i], factor_noise,
				      list_filters_sample,list_downsample_filters_sample, size_list_sample,
				      size_list_downsample_sample, list_filters_texture_analysis,
				      list_downsample_filters_texture, list_filters_texture_synthesis,
				      list_mask_filters, size_list_texture,
				      size_list_downsample_texture, edge_handling);

  /* Make the invert change of basis PCA --> RGB */
  if(color == 3)
  {
     for(unsigned int i=0;i<3;i++)
     {
      for(unsigned int j=0;j<nxout*nyout;j++)
      {
       data_out_rgb[i][j]  =  data_in_rgb_mean[i];
       for(int k=0;k<3; k++) data_out_rgb[i][j] += M_eigen_vectors[i][k]*data_out_rgb_pca[k][j];
       data_out_rgb[i][j] = floor(data_out_rgb[i][j]);
      }
     }
  }
  else
  {
    for(unsigned int j=0;j<nxout*nyout;j++)
      {
       data_out_rgb[0][j]  =  floor(data_out_rgb_pca[0][j]);
      }
  }

 /* Smooth component part */
 if(smooth == 1)
 {
   		     /* Make a zoom if there is an extension */
		       if (nxin == nxout && nyin == nyout)
		        {
			      for(int i=0;i<color;i++)
			      {
				for(unsigned int j=0;j<nxout*nyout;j++) data_out_rgb[i][j]+=
				  data_in_rgb[i][j]-data_in_rgb_per[i][j];
			      }
		        }
		       else if( nxout == (nxout/nxin)*nxin && nyout == (nyout/nyin)*nyin )
		        {
			    float *data_in_smooth = (float *) malloc(color * nxin * nyin * sizeof(float));
			    float *data_in_smooth_rgb[color];
				for(int d=0;d<color;d++)	data_in_smooth_rgb[d] = data_in_smooth+d*nxout*nyout;

			    float *data_zoom_smooth = (float *) malloc(color * nxout * nyout * sizeof(float));
			    float *data_zoom_smooth_rgb[color];
				for(int d=0;d<color; d++)	data_zoom_smooth_rgb[d] = data_zoom_smooth +d*nxout*nyout;

			    for(int i=0;i<color;i++)
				{
				  for(unsigned int j=0;j<nxout*nyout;j++) data_in_smooth_rgb[i][j] =
				    data_in_rgb[i][j]-data_in_rgb_per[i][j];
				}

			    for(int i=0;i<color;i++)
			    {
			     zoom_bilin(data_zoom_smooth_rgb[i], nxout, nyout,data_in_smooth_rgb[i],  nxin, nyin);
			     for(unsigned int j=0;j<nxout*nyout;j++) data_out_rgb[i][j] += data_zoom_smooth_rgb[i][j];
			    }

			 free(data_in_smooth);
			 free(data_zoom_smooth);
		        }
 }

  /* free */
    free(size_list_sample_vector);
    free(size_list_downsample_sample_vector);
    free(size_list_texture_vector);
    free(size_list_downsample_texture_vector);
    free(data_in_pca);
    free(data_in_per);
    free(data_out_pca);

       for(int i=0;i<2+N_pyr*N_steer;i++) free(list_filters_sample[i]);
       for(int i=0;i<1+N_pyr;i++) free(list_downsample_filters_sample[i]);
       for(int i=0;i<2+N_pyr*N_steer;i++) {
	      free(list_filters_texture_analysis[i]);
	      free(list_filters_texture_synthesis[i]);
         }
       for(int i=0;i<1+N_pyr;i++) free(list_downsample_filters_texture[i]);
       for(int i=0;i<N_pyr*N_steer;i++) free(list_mask_filters[i]);

}